# Copyright (c) 2022 Marcus Schäfer.  All rights reserved.
#
# This file is part of kiwi.
#
# kiwi is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# kiwi is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with kiwi.  If not, see <http://www.gnu.org/licenses/>
#
import os
import logging
from typing import (
    List, Dict, Optional, Union
)

# project
from kiwi.defaults import Defaults
from kiwi.mount_manager import MountManager
from kiwi.storage.disk import ptable_entry_type
from kiwi.volume_manager.base import VolumeManagerBase
from kiwi.filesystem.base import FileSystemBase

log = logging.getLogger('kiwi')


class ImageSystem:
    """
    **Access the target image from the block layer**
    """
    def __init__(
        self, device_map: Dict, root_dir: str,
        volume_manager: Optional[Union[FileSystemBase, VolumeManagerBase]] = None,
        partitions: Dict[str, ptable_entry_type] = {}
    ) -> None:
        """
        Construct a new ImageSystem object

        :param dict device_map: Block device map
        :param str root_dir: Path to unpacked image root tree
        :param VolumeManagerBase volume_manager:
            Optional VolumeManager instance
        """
        self.arch = Defaults.get_platform_name()
        self.device_map = device_map
        self.root_dir = root_dir
        self.volume_manager = volume_manager
        self.partitions = partitions
        self.mount_list: List[MountManager] = []
        self.root_mount_mountpoint = ''

    def __enter__(self):
        return self

    def mountpoint(self) -> str:
        """
        Return image root mountpoint

        :return: mountpoint path or empty string

        :rtype: str
        """
        return self.root_mount_mountpoint

    def mount(self) -> None:
        """
        Mount image system from current block layers
        """
        # mount root boot and efi devices as they are present
        (root_device, boot_device, efi_device) = self._setup_device_names()
        if self.volume_manager:
            self.volume_manager.mount_volumes()
            self.root_mount_mountpoint = \
                self.volume_manager.get_mountpoint() or ''
        else:
            root_mount = MountManager(device=root_device)
            self.mount_list.append(root_mount)
            root_mount.mount()
            self.root_mount_mountpoint = root_mount.mountpoint

        if 's390' in self.arch:
            boot_mount = MountManager(
                device=boot_device, mountpoint=os.path.join(
                    self.root_mount_mountpoint, 'boot', 'zipl'
                )
            )
        else:
            boot_mount = MountManager(
                device=boot_device, mountpoint=os.path.join(
                    self.root_mount_mountpoint, 'boot'
                )
            )
        if efi_device:
            efi_mount = MountManager(
                device=efi_device, mountpoint=os.path.join(
                    self.root_mount_mountpoint, 'boot', 'efi'
                )
            )
        if root_device != boot_device:
            self.mount_list.append(boot_mount)
            boot_mount.mount()

        if efi_device:
            self.mount_list.append(efi_mount)
            efi_mount.mount()

        if self.partitions:
            for map_name in sorted(self.partitions.keys()):
                if map_name in self.device_map:
                    partition_mount = MountManager(
                        device=self.device_map[map_name].get_device(),
                        mountpoint=os.path.join(
                            self.root_mount_mountpoint,
                            self.partitions[map_name].mountpoint.lstrip(os.sep)
                        )
                    )
                    self.mount_list.append(partition_mount)
                    partition_mount.mount()

        # bind mount /image from unpacked root to get access to e.g scripts
        image_mount = MountManager(
            device=os.path.join(self.root_dir, 'image'),
            mountpoint=os.path.join(
                self.root_mount_mountpoint, 'image'
            )
        )
        self.mount_list.append(image_mount)
        image_mount.bind_mount()

        # mount tmp as tmpfs
        tmp_mount = MountManager(
            device='tmpfs', mountpoint=os.path.join(
                self.root_mount_mountpoint, 'tmp'
            )
        )
        self.mount_list.append(tmp_mount)
        tmp_mount.tmpfs_mount()

        # mount var/tmp as tmpfs
        var_tmp_mount = MountManager(
            device='tmpfs', mountpoint=os.path.join(
                self.root_mount_mountpoint, 'var', 'tmp'
            )
        )
        self.mount_list.append(var_tmp_mount)
        var_tmp_mount.tmpfs_mount()

        # mount kernel interfaces as bind
        for location in ('proc', 'sys', 'dev'):
            shared_mount = MountManager(
                device=os.path.join('/', location), mountpoint=os.path.join(
                    self.root_mount_mountpoint, location
                )
            )
            self.mount_list.append(shared_mount)
            shared_mount.bind_mount()

    def umount(self) -> None:
        """
        Umount all elements of mount_list in reverse order
        """
        for mount in reversed(self.mount_list):
            if mount.is_mounted():
                mount.umount()
        if self.volume_manager:
            self.volume_manager.umount_volumes()

    def _setup_device_names(self) -> tuple:
        root_device = self.device_map['root'].get_device()
        boot_device = root_device
        efi_device = ''
        if 'readonly' in self.device_map:
            root_device = self.device_map['readonly'].get_device()
            boot_device = root_device
        if 'boot' in self.device_map:
            boot_device = self.device_map['boot'].get_device()
        if 'efi' in self.device_map:
            efi_device = self.device_map['efi'].get_device()
        return (root_device, boot_device, efi_device)

    def __exit__(self, exc_type, exc_value, traceback):
        self.umount()



# Copyright (c) 2024 SUSE Software Solutions Germany GmbH.  All rights reserved.
#
# This file is part of kiwi.
#
# kiwi is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# kiwi is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with kiwi.  If not, see <http://www.gnu.org/licenses/>
#
from typing import List

# project
from kiwi.filesystem.base import FileSystemBase
from kiwi.command import Command


class FileSystemEroFs(FileSystemBase):
    """
    **Implements creation of erofs filesystem**
    """
    def create_on_file(
        self, filename, label: str = None, exclude: List[str] = None
    ):
        """
        Create erofs filesystem from data tree

        :param string filename: result file path name
        :param string label: volume label
        :param list exclude: list of exclude dirs/files
        """
        self.filename = filename
        exclude_options = []
        compression = self.custom_args.get('compression')
        call_args = self.custom_args['create_options'].copy()
        if compression:
            call_args.append('-z')
            call_args.append(compression)

        if exclude:
            for item in exclude:
                # item is a glob, but erofs requires a POSIX extended regex.
                # Translate the glob to a regex for correct behaviour.
                #
                # We can't use fnmatch.translate, as that produces a Python regex.
                as_regex = '^' + item.replace('*', '.*') + '$'
                exclude_options.append(f'--exclude-regex={as_regex}')

        if label:
            call_args.append('-L')
            call_args.append(label)

        Command.run(
            [
                'mkfs.erofs'
            ] + call_args + exclude_options + [
                self.filename, self.root_dir
            ]
        )



# Copyright (c) 2026 SUSE LLC.  All rights reserved.
#
# This file is part of kiwi.
#
# kiwi is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# kiwi is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with kiwi.  If not, see <http://www.gnu.org/licenses/>
#
from string import Template
from textwrap import dedent


class OvaComposeTemplate:
    """
    OVA compose settings template
    """
    def __init__(self):
        self.cr = '\n'

        self.header = dedent('''
            system:
                name: "${display_name}"
                type: vmx-20
                os_vmw: "${guest_os}"
                firmware: "${firmware}"
                secure_boot: ${secure_boot}
        ''').strip() + self.cr

        self.hardware = dedent('''
            hardware:
                cpus: ${number_of_cpus}
                memory: ${memory_size}
                sata1:
                    type: sata_controller
                scsi1:
                    type: scsi_controller
                nvme1:
                    type: nvme_controller
                cdrom1:
                    type: cd_drive
                    parent: sata1
                rootdisk:
                    type: hard_disk
                    parent: nvme1
                    disk_image: "${vmdk_file}"
                usb2:
                    type: usb_controller
                usb3:
                    type: usb3_controller
                ethernet1:
                    type: ethernet
                    subtype: VmxNet3
                    network: vm_network
                videocard1:
                    type: video_card
                vmci1:
                    type: vmci
        ''').strip() + self.cr

        self.network = dedent('''
            networks:
                vm_network:
                    name: "VM Network"
                    description: "The VM Network network"
        ''').strip() + self.cr

    def get_template(self, hardware_setup: bool = False):
        """
        Set data for the ova-compose template

        :param hardware_setup: if True, include hardware section

        :rtype: Template
        """
        template_data = self.header
        template_data += self.network
        if hardware_setup:
            template_data += self.hardware

        return Template(template_data)



# Copyright (c) 2015 SUSE Linux GmbH.  All rights reserved.
#
# This file is part of kiwi.
#
# kiwi is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# kiwi is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with kiwi.  If not, see <http://www.gnu.org/licenses/>
#
import logging
import simplejson
import pickle
import os
from typing import (
    Dict, NamedTuple, TypeVar, Any, Optional
)

# project
from kiwi.xml_state import XMLState
from kiwi.exceptions import (
    KiwiResultError
)

log = logging.getLogger('kiwi')

# must be global to allow pickle to find it
result_file_type = NamedTuple(
    'result_file_type', [
        ('filename', str),
        ('use_for_bundle', bool),
        ('compress', bool),
        ('shasum', bool)
    ]
)
result_name_tags = NamedTuple(
    'result_name_tags', [
        ('N', str),  # image name
        ('P', str),  # concatenated profile name (_)
        ('A', str),  # architecture name
        ('I', str),  # custom ID setting
        ('T', str),  # image type name
        ('M', str),  # Major version number
        ('m', str),  # Minor version number
        ('p', str),  # Patch version number
        ('v', str)   # Version string
    ]
)

result_type = TypeVar('result_type', bound='Result')


class Result:
    """
    **Collect image building results**

    :param dict result_files: dict of result files
    :param object class_version: :class:`Result` class version
    :param object xml_state: instance of :class:`XMLState`
    """
    def __init__(self, xml_state: XMLState):
        self.result_files: Dict[str, Any] = {}

        # Instances of this class are stored as result reference.
        # In order to handle class format changes any instance
        # provides a version information
        self.class_version: int = 1

        self.xml_state = xml_state

    def add_bundle_format(self, pattern: str):
        version_string = self.xml_state.get_image_version()
        if '.' in version_string:
            (major, minor, patch) = version_string.split(sep='.', maxsplit=2)
        else:
            (major, minor, patch) = (version_string, '', '')
        self.name_tags = result_name_tags(
            N=self.xml_state.xml_data.get_name(),
            P='_'.join(
                self.xml_state.profiles
            ) if self.xml_state.profiles else '',
            A=self.xml_state.host_architecture,
            I='',
            T=self.xml_state.get_build_type_name(),
            M=major,
            m=minor,
            p=patch,
            v=version_string
        )
        self.result_files['bundle_format'] = {
            'pattern': pattern,
            'tags': self.name_tags
        }

    def get_name_for_pattern(self, command_args: Dict[str, str] = {}) -> str:
        bundle_format = self.result_files.get('bundle_format')
        bundle_file_format_name = ''
        if bundle_format:
            tags = bundle_format['tags']
            bundle_file_format_name = bundle_format['pattern']
            # Insert image name
            bundle_file_format_name = bundle_file_format_name.replace(
                '%N', tags.N
            )
            # Insert Concatenated profile name (_)
            bundle_file_format_name = bundle_file_format_name.replace(
                '%P', tags.P
            )
            # Insert Architecture name
            bundle_file_format_name = bundle_file_format_name.replace(
                '%A', tags.A
            )
            # Insert Image build type name
            bundle_file_format_name = bundle_file_format_name.replace(
                '%T', tags.T
            )
            # Insert Image Major version number
            bundle_file_format_name = bundle_file_format_name.replace(
                '%M', format(tags.M)
            )
            # Insert Image Minor version number
            bundle_file_format_name = bundle_file_format_name.replace(
                '%m', format(tags.m)
            )
            # Insert Image Patch version number
            bundle_file_format_name = bundle_file_format_name.replace(
                '%p', format(tags.p)
            )
            # Insert Version string
            bundle_file_format_name = bundle_file_format_name.replace(
                '%v', format(tags.v)
            )
            # Insert Bundle ID
            if command_args.get('--id'):
                bundle_file_format_name = bundle_file_format_name.replace(
                    '%I', command_args['--id']
                )
        return bundle_file_format_name

    def add(
        self, key: str, filename: str, use_for_bundle: bool = True,
        compress: bool = False, shasum: bool = True
    ) -> None:
        """
        Add result tuple to result_files list

        :param str key: name
        :param str filename: file path name
        :param bool use_for_bundle: use when bundling results true|false
        :param bool compress: compress when bundling true|false
        :param bool shasum: create shasum when bundling true|false
        """
        if key and filename:
            self.result_files[key] = result_file_type(
                filename=filename,
                use_for_bundle=use_for_bundle,
                compress=compress,
                shasum=shasum
            )

    def get_results(self) -> Dict[str, result_file_type]:
        """
        Current list of result tuples
        """
        return self.result_files

    def print_results(self) -> None:
        """
        Print results human readable
        """
        if self.result_files:
            log.info('Result files:')
            for key, value in sorted(list(self.result_files.items())):
                if key == 'bundle_format':
                    log.info('--> {0}: {1}'.format(key, value.get('pattern')))
                else:
                    log.info('--> {0}: {1}'.format(key, value.filename))

    def dump(self, filename: str, portable: bool = True) -> None:
        """
        Picke dump this instance to a file

        :param str filename: file path name
        :param bool portable:
            If set to true also create a .json formatted variant
            of the dump file which contains the elements of this
            instance that could be expressed in a portable json
            document. Default is set to: True

        :raises KiwiResultError: if pickle fails to dump :class:`Result`
            instance
        """
        try:
            with open(filename, 'wb') as result:
                pickle.dump(self, result)
            if portable:
                with open(filename + '.json', 'w') as result_portable:
                    result_portable.write(
                        simplejson.dumps(
                            self.result_files, sort_keys=True, indent=4
                        ) + os.linesep
                    )
        except Exception as e:
            raise KiwiResultError(
                f'Failed to pickle dump results: {format(e)}'
            )

    @staticmethod
    def load(filename: str) -> result_type:  # type: ignore
        """
        Load pickle dumped filename into a Result instance

        :param str filename: file path name

        :raises KiwiResultError: if filename does not exist or pickle fails
            to load filename
        """
        if not os.path.exists(filename):
            raise KiwiResultError(
                f'No result information {filename} found'
            )
        try:
            with open(filename, 'rb') as result:
                return pickle.load(result)
        except Exception as e:
            raise KiwiResultError(
                f'Failed to pickle load results: {type(e).__name__}'
            )

    @staticmethod
    def verify_image_size(size_limit: Optional[int], filename: str) -> None:
        """
        Verifies the given image file does not exceed the size limit.
        Throws an exception if the limit is exceeded. If the size limit
        is set to None no verification is done.

        :param int size_limit: The size limit for filename in bytes.
        :param str filename: File to verify.

        :raises KiwiResultError: if filename exceeds the size limit
        """
        if size_limit is not None:
            if os.path.getsize(filename) > size_limit:
                raise KiwiResultError(
                    'Build constraint failed: {0} is bigger than {1}'
                    .format(filename, size_limit)
                )
